/*	voodooacpi.c - M.A.R-T>Y. ACPI control daemon written in C
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include <poll.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>

#include <errno.h>

#include <limits.h>

#ifndef LIBMMP_AUTHOR
	#include "libmmp.h"
#endif

#ifndef LIBMARTY_AUTHOR
	#include "libmarty.h"
#endif

#ifndef MENDPOINTS_H
	#include "mendpoints.h"
#endif

const int TARGET_PROTOCOL[3] = {2, 0, 0};
const int INSTANCE_VERSION[3] = {3, 1, 2};

const char INSTANCE_NAME[] = "m-cvoodooacpi";
const char INSTANCE_TITLE[] = "VoodooACPI Control Daemon";

struct WakeEntry {
	char *hostname;
	char *mac;
};

struct MartyClientEndpoint Endpoints[] = {
	{
		.endpoint = VOODOO_EPT_SLEEP,
		.name = "sleep",
		.min = 5,
		.max = 5,
		.isAscii = false
	},
	{
		.endpoint = VOODOO_EPT_SHUTDOWN,
		.name = "shutdown",
		.min = 5,
		.max = 5,
		.isAscii = false
	},
	{
		.endpoint = VOODOO_EPT_WAKE,
		.name = "wake",
		.min = 6,
		.max = 6,
		.isAscii = true
	}
};

int MaxEndpoint = sizeof(Endpoints) / sizeof(Endpoints[0]);

struct WakeEntry Hosts[] = {
	{
		.hostname = "sunyshore",
		.mac = "00:00:00:00:00:00"
	},
	{
		.hostname = "blackthorn",
		.mac = "00:00:00:00:00:00"
	}
};

int MaxHost = sizeof(Hosts) / sizeof(Hosts[0]);

int main(int argc, char **argv) {
	pthread_t networkThread;
	struct MartyClientData data;
	struct MMStatus status;

	int bytesRead = 0;
	struct MMPacket txPacket;
	struct MMPacket *rxPacket = NULL;

	mmp_arg applicationStatus = MMP_APPLICATION_SUCCESS;

	char hostname[_POSIX_HOST_NAME_MAX];

	int endpointConversion = -1;
	struct MartyClientEndpoint *endpoint = NULL;

	struct MartyMessageHandle *deliveryHandle = NULL;

	struct WakeEntry *host = NULL;

	char *execCommand = NULL;
	char **execArgs = NULL;

	char *sleepArgs[] = {
		"loginctl",
		"suspend",
		NULL
	};

	char *shutdownArgs[] = {
		"poweroff",
		NULL
	};

	char *wakeArgs[] = {
		"wakeonlan",
		"-i",
		"192.168.1.255",
		NULL,
		NULL
	};

	printf("%s[%s]%s Starting hoff.industries %s - %s v%d.%d.%d\n",
		   LoggerColours[LOG_INFO], MartyFlavour, ClearColour, INSTANCE_TITLE,
		INSTANCE_NAME, INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2]);

	if (signal(SIGCHLD, SIG_IGN) == SIG_ERR) {
		marty_part_log("Signal ignore registration failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_SIGN);
	};

	if (marty_client_data_factory(&data) == -1) {
		marty_part_log("Client data init failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_CDATA);
	}

	mmstatus_factory(&status);
	mmpacket_factory(&txPacket, true);

	data.versionString = alloc_marty_version_string(INSTANCE_VERSION, INSTANCE_NAME);
	if (data.versionString == NULL) {
		marty_part_log("Version string allocation failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_VSTRING);
	}

	if (gethostname(hostname, _POSIX_HOST_NAME_MAX) == -1) {
		marty_part_log("Hostname get failure!\n", LOG_CRITICAL);
		exit(EXIT_ERR_HOSTNAME);
	}

	data.connType = CONN_UNIX;
	data.initParameters.auth = "";
	data.initParameters.priority = 0;
	data.initParameters.host = hostname;
	data.initParameters.name = "voodooacpi";

	if (spawn_marty_socket(&data) != MMP_ERR_NONE) {
		marty_full_log("Socket spawn failure!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}

	if (marty_init(&data) != MMP_ERR_NONE) {
		marty_full_log("MMP initialisation failure!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_INIT);
	};

	marty_full_log("MMP initialisation success!\n", LOG_STARTUP, &data);

	if (pthread_create(&networkThread, NULL, marty_network_loop, (void *)&data) != 0) {

		marty_full_log("Error spawning network thread!\n", LOG_CRITICAL, &data);
		exit(EXIT_ERR_SPAWN);
	}
	marty_full_log("Network loop started.\n", LOG_STARTUP, &data);

	while (true) {

		mmpacket_reset(&txPacket, true);
		applicationStatus = MMP_APPLICATION_SUCCESS;

		if (rxPacket != NULL) {
			mmpacket_reset(rxPacket, true);
			free(rxPacket);
		}

		rxPacket = NULL;

		if (poll(data.pfds, data.nfds, -1) == -1) {
			marty_full_log("Error polling inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_POLL);
		}

		if (data.pfds[ITC_DELIVERY_PIPE].revents != 0) {
			// check APPLICATION events

			bytesRead = read(data.netDeliveryPipe[0], &deliveryHandle,
				sizeof(struct MartyMessageHandle *));


			if (bytesRead == 0 || bytesRead == -1) {
				marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
				exit(EXIT_ERR_ITC);
			}

			if (deliveryHandle == NULL) {
				// if ITC pipe sent -2 means reconnect event occurred.
			}

			continue;
		}

		if (data.pfds[ITC_MESSAGE_PIPE].revents == 0) {
			// since MESSAGE events are last on our list to check
			// we can continue if we don't need to handle
			continue;
		}

		bytesRead = read(data.netMessagePipe[0], &rxPacket,
			sizeof(struct MMPacket *));

		if (bytesRead == 0 || bytesRead == -1) {
			marty_full_log("Error reading inter-thread pipes!\n", LOG_CRITICAL, &data);
			exit(EXIT_ERR_ITC);
		}

		if (rxPacket == NULL) {
			marty_full_log("Received NULL pointer from ITC pipe!\n", LOG_WARNING, &data);
			continue;
		}

		endpointConversion = rxPacket->payloads[PAYL_IDX_ENDPOINT][0];

		if (endpointConversion > MaxEndpoint || endpointConversion < 0) {
			marty_full_log("Requested endpoint outside my range!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_ENDPOINT;

			goto JMP_TX_STATUS;
		}

		endpoint = &Endpoints[endpointConversion];

		if (rxPacket->payloadLen < endpoint->min) {
			marty_full_log("Request payload count less than expected!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_TOO_FEW;

			goto JMP_TX_STATUS;
		}

		if (rxPacket->payloadLen > endpoint->max) {
			marty_full_log("Request payload count more than expected!\n", LOG_WARNING, &data);
			applicationStatus = MMP_APPLICATION_TOO_MANY;

			goto JMP_TX_STATUS;
		}

		switch (endpointConversion) {
			case VOODOO_EPT_SLEEP:
				execCommand = "loginctl";
				execArgs = sleepArgs;
				break;

			case VOODOO_EPT_SHUTDOWN:
				execCommand = "poweroff";
				execArgs = shutdownArgs;
				break;

			case VOODOO_EPT_WAKE:
				execCommand = "wakeonlan";
				execArgs = wakeArgs;

				rxPacket->payloads[5][mmpayloadlen(rxPacket->payloads[5])] = '\0';

				for (int i = 0; i < MaxHost; i++) {
					if (!strcmp((char *)rxPacket->payloads[5], Hosts[i].hostname)) {
						host = &Hosts[i];
						break;
					}
				}

				if (host == NULL) {
					marty_full_log("Requested hostname not recorded!\n", LOG_ERROR, &data);
					applicationStatus = MMP_APPLICATION_EXECUTE;

					goto JMP_TX_STATUS;
				}

				execArgs[3] = host->mac;

				host = NULL;
				break;

			default:
				marty_full_log("Not handling this endpoint.\n", LOG_WARNING, &data);
				applicationStatus = MMP_APPLICATION_ENDPOINT;

				goto JMP_TX_STATUS;
		}

		if (fork() == 0 && execvp(execCommand, execArgs) == -1) {

			if (errno != ENOENT) {
				marty_full_log("Child process execution failure!\n", LOG_ERROR, &data);
				exit(EXIT_ERR_CHILD);
			}

			marty_full_log("Child process execution failure!\n", LOG_ERROR, &data);

			// we'll try invoking systemd explicitly
			if (endpointConversion == VOODOO_EPT_SLEEP) {
				execCommand = "systemctl";
				execArgs[0] = "systemctl";
			}

			if (endpointConversion == VOODOO_EPT_SHUTDOWN) {
				execCommand = "systemctl";

				execArgs = sleepArgs;
				execArgs[0] = "systemctl";
				execArgs[1] = "poweroff";
			}

			if (execvp(execCommand, execArgs) == -1) {
				marty_full_log("Child process execution failure... again!\n", LOG_ERROR, &data);
			}

			exit(EXIT_ERR_CHILD);
		}

		JMP_TX_STATUS:

		marty_set_application(&txPacket, &status, rxPacket->payloads[0], applicationStatus);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't set autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		mmp_serialize(&txPacket, &status);
		if (status.error != MMP_ERR_NONE) {
			marty_full_log("Couldn't serialize autopacket!\n", LOG_ERROR, &data);
			continue;
		}

		marty_net_write(&txPacket, &status, &data);
	}
}
